import yaml
from glob import glob
import os
import shutil

from .utils import run_and_save_logs
from .utils_ros import parse_tests_results
from .parameters import TMP_SCENARIO_PARAMS_YAML


def generate_scenario_parameter_output(params: dict, param_file: str):
    """
    Store `params` in `param_file` and convert to ros2 param file nested format,
    to be used by the launch file
    """
    content = {}
    for k, v in params.items():
        try:
            node, pname = k.split("/")
        except Exception:
            print("ensure params are in the format `{node_name}/{param_name}`")
            return
        if node not in content:
            content[node] = {"ros__parameters": {}}
        content[node]["ros__parameters"][pname] = v
    with open(param_file, "w") as f:
        yaml.dump(content, f)


def run_ros2_tests(run):
    scenario = run.params
    # TODO: HOW TO ADD  NODE to launch
    # TODO: set params from conf
    # TODO: get params to log
    # TODO: where is the rosbag
    if "params" in run.params:
        # note: fixed filename will lead concurent executions to overwrite each other
        generate_scenario_parameter_output(
            run.params["params"], TMP_SCENARIO_PARAMS_YAML
        )

    preexisting_rosbags = glob("rosbag2*")
    test_result_file_path = f"{run.output_path}/tests_junit.xml"
    launch_arguments = [
        f"{k}:={v}" for k, v in run.params.get("launch_arguments", {}).items()
    ]
    command = [
        "launch_test",
        "--junit-xml",
        test_result_file_path,
        scenario["ros_testfile"],
    ] + launch_arguments

    # save ROS logs in the output dir
    ros_log_dir = os.path.join(run.output_path, "ros_logs")

    # Main: test execution
    # shell=True required to support command list items that are strings with spaces
    # (this way, scenario["ros_testfile"] can be either a path to the launch file or '<package_name> <launch_name>')
    return_code = run_and_save_logs(
        " ".join(command),
        shell=True,
        executable="/bin/bash",
        env={
            **os.environ,
            **{
                "ROS_LOG_DIR": ros_log_dir,
                "ARTEFACTS_SCENARIO_PARAMS_FILE": TMP_SCENARIO_PARAMS_YAML,
            },
        },
        output_path=os.path.join(run.output_path, "test_process_log.txt"),
    )
    if return_code == 2:
        raise Exception(
            f"Running {scenario['ros_testfile']} failed. Please check that the launch file exists."
        )

    # zip ROS logs and delete the original folder
    if os.path.exists(ros_log_dir):
        shutil.make_archive(ros_log_dir, "zip", ros_log_dir)
        shutil.rmtree(ros_log_dir)

    # parse xml generated by launch_test
    results, success = parse_tests_results(test_result_file_path)
    if success is None:
        run.log_tests_results(results, False)
        return results, success
    run.log_artifacts(run.output_path)

    # upload any additional files in the folders specified by the user in artefacts.yaml
    for output in scenario.get("output_dirs", []):
        run.log_artifacts(output)

    # check if any rosbag was created
    rosbags = glob("rosbag2*")
    new_rosbags = set(rosbags).difference(set(preexisting_rosbags))
    from artefacts.cli.bagparser import BagFileParser

    if len(new_rosbags) > 0:
        rosbag_path = new_rosbags.pop()
        run.log_artifacts(rosbag_path, "rosbag")
        if "metrics" in run.params:
            # TODO should go inside BagFileParser?
            db_files = glob(f"{rosbag_path}/*.mcap")  # Ros2 Default
            if not db_files:
                db_files = glob(f"{rosbag_path}/*.db3")  # Legacy
            if not db_files:
                raise FileNotFoundError(
                    "No .mcap or .db3 files found in the specified path."
                )
            db_file = db_files[0]
            bag = BagFileParser(db_file)
            for metric in run.params["metrics"]:
                try:
                    last_value = bag.get_last_message(metric)[1].data
                    run.log_metric(metric, last_value)
                except KeyError:
                    print(f"Metric {metric} not found in rosbag, skipping.")
                except TypeError or IndexError:
                    print(
                        f"Metric {metric} not found. Is it being published?. Skipping."
                    )

    run.log_tests_results(results, success)
    return results, success
